Set up directories
  $ . $TESTDIR/setup.sh 

Set up the executable: toAfg.
  $ EXEC=$TESTDIR/../toAfg

test basic FASTA use case
  $ $EXEC $DATDIR/read.fasta output.afg -noSplitSubreads
  $ head output.afg
  {UNV
  iid:1
  com:
  generated by AfgBasWriter
  Mon Jun 28 14:43:52 2010
  .
  }
  {LIB
  iid:1
  {DST
  $ tail output.afg
  qlt:
  55555555555555555555555555555555555555555555555555555555555555555555555555555555
  55555555555555555555
  .
  }
  {FRG
  iid:24
  lib:1
  typ:I
  }

test uniform QV use case
  $ $EXEC $DATDIR/test_toafg.fastq output.afg -noSplitSubreads -uniformQV 3
  $ head output.afg
  {UNV
  iid:1
  com:
  generated by AfgBasWriter
  Mon Jun 28 14:43:52 2010
  .
  }
  {LIB
  iid:1
  {DST

test uniform QV FASTA use case
  $ $EXEC $DATDIR/read.fasta output.afg -noSplitSubreads -uniformQV 7
  $ head output.afg
  {UNV
  iid:1
  com:
  generated by AfgBasWriter
  Mon Jun 28 14:43:52 2010
  .
  }
  {LIB
  iid:1
  {DST
  $ tail output.afg
  qlt:
  77777777777777777777777777777777777777777777777777777777777777777777777777777777
  77777777777777777777
  .
  }
  {FRG
  iid:24
  lib:1
  typ:I
  }

